/****************************************************************************
**
** This file is part of a Qt Solutions component.
** 
** Copyright (c) 2009 Nokia Corporation and/or its subsidiary(-ies).
** 
** Contact:  Qt Software Information (qt-info@nokia.com)
** 
** Commercial Usage  
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Solutions Commercial License Agreement provided
** with the Software or, alternatively, in accordance with the terms
** contained in a written agreement between you and Nokia.
** 
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
** 
** In addition, as a special exception, Nokia gives you certain
** additional rights. These rights are described in the Nokia Qt LGPL
** Exception version 1.0, included in the file LGPL_EXCEPTION.txt in this
** package.
** 
** GNU General Public License Usage 
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
** 
** Please note Third Party Software included with Qt Solutions may impose
** additional restrictions and it is the user's responsibility to ensure
** that they have met the licensing requirements of the GPL, LGPL, or Qt
** Solutions Commercial license and the relevant license of the Third
** Party Software they are using.
** 
** If you are unsure which license is appropriate for your use, please
** contact the sales department at qt-sales@nokia.com.
** 
****************************************************************************/



//Own
#include "states.h"
#include "graphicsscene.h"
#include "boat.h"
#include "submarine.h"
#include "torpedo.h"
#include "animationmanager.h"

//Qt
#include <QMessageBox>
#include <QGraphicsView>
#if defined(QT_EXPERIMENTAL_SOLUTION)
# include "qtstatemachine.h"
# include "qtkeyeventtransition.h"
#else
# include <QtCore/QtStateMachine>
# include <QtGui/QtKeyEventTransition>
#endif

PlayState::PlayState(GraphicsScene *scene, QtState *parent) : QtState(parent), scene(scene), machine(0), currentLevel(1)
{
}

PlayState::~PlayState()
{
}

void PlayState::onEntry()
{
    //We are now playing?
    if (machine) {
        machine->stop();
        scene->clearScene();
        delete machine;
    }

    machine = new QtStateMachine(this);

    //This state is when player is playing
    QtState *playState = new QtState(machine->rootState());

    initializeLevel();

    //This state is when the game is paused
    PauseState *pauseState = new PauseState(scene,machine->rootState());

    //We have one view, it receive the key press event
    QtKeyEventTransition *pressPplay = new QtKeyEventTransition(scene->views().at(0),QEvent::KeyPress,Qt::Key_P);
    QtKeyEventTransition *pressPpause = new QtKeyEventTransition(scene->views().at(0),QEvent::KeyPress,Qt::Key_P);

    //Pause "P" is triggered, the player pause the game
    playState->addTransition(pressPplay, pauseState);

    //To get back playing when the game has been paused
    pauseState->addTransition(pressPpause, playState);

    //This state is when player have lost
    LostState *lostState = new LostState(scene,machine->rootState());

    //This state is when player have won
    WinState *winState = new WinState(scene,machine->rootState());

    //The boat has been destroyed then the game is finished
    playState->addTransition(scene->boat, SIGNAL(boatExecutionFinished()),lostState);

    //This transition check if we won or not
    WinTransition *winTransition = new WinTransition(scene);

    //The boat has been destroyed then the game is finished
    playState->addTransition(winTransition,winState);

    //We start playing!!!
    machine->setInitialState(playState);

    machine->start();
}

void PlayState::initializeLevel()
{
    scene->boat = new Boat();
    scene->addItem(scene->boat);
    scene->setFocusItem(scene->boat,Qt::OtherFocusReason);
    scene->boat->setPos(scene->width()/2, scene->sealLevel() - scene->boat->size().height());

    for (int i = 0 ; i <= 5 ; ++i) {
        SubMarine *sub = new SubMarine();
        scene->addItem(sub);
        int random = (qrand() % 15 + 1);
        qreal x = random == 13 || random == 5 ? 0 : scene->width() - sub->size().width();
        qreal y = scene->height() -(qrand() % 150 + 1) - sub->size().height();
        sub->setPos(x,y);
        sub->setCurrentDirection(x == 0 ? SubMarine::Right : SubMarine::Left);
        sub->setCurrentSpeed(qrand() % 3 + 1);
    }
}

/** Pause State */
PauseState::PauseState(GraphicsScene *scene, QtState *parent) : QtState(parent),scene(scene)
{
}
void PauseState::onEntry()
{
    AnimationManager::self()->pauseAll();
    scene->boat->setEnabled(false);
}
void PauseState::onExit()
{
    AnimationManager::self()->resumeAll();
    scene->boat->setEnabled(true);
    scene->boat->setFocus();
}

/** Lost State */
LostState::LostState(GraphicsScene *scene, QtState *parent) : QtState(parent),scene(scene)
{
}

void LostState::onEntry()
{
    //We clear the scene
    scene->clearScene();

    //we have only one view
    QMessageBox::information(scene->views().at(0),"You loose","You loose!!!!!!!!!!");
}

/** Win State */
WinState::WinState(GraphicsScene *scene, QtState *parent) : QtState(parent),scene(scene)
{
}

void WinState::onEntry()
{
    //We clear the scene
    scene->clearScene();

    //we have only one view
    QMessageBox::information(scene->views().at(0),"You win","You win!!!!!!!!!!");
}

/** Win transition */
WinTransition::WinTransition(GraphicsScene *scene) : QtSignalTransition(scene,SIGNAL(subMarineDestroyed()))
{
    this->scene = scene;
}

bool WinTransition::testCondition(QEvent *event) const
{
    Q_UNUSED(event)
    if (scene->remainingSubMarines() == 0)
        return true;
    else
        return false;
}
